/*=============================================================================
vitalnix/accdb - User Management Suite - Unified Account Database
accdb_shadow.c - Shadow back-end module
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2002 - 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <sys/file.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libHX.h>
#include <accdb.h>
#include <accdb_int.h>
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

enum {
  ID_USER    = 1 << 0,
  ID_GROUP   = 1 << 1,
  IDF_SYSTEM = 1 << 2,
};

enum lock_request {
  LOCK_OFF = 0,
  LOCK_ON,
};

struct state {
  enum accdb_sysctl_flags flags;
  long uid_min, uid_max, gid_min, gid_max;

  FILE *fp_pw, *fp_sh, *fp_gr, *fp_gs;
  byte chg_pw, chg_sh, chg_gr, chg_gs;
  byte lockstate;

  struct HXlld *user, *group;
  struct HXlld_node *utrav, *gtrav;
};

extern int q_open(void **);
extern int q_usertrav(void *, struct accdb_user *);
extern int q_userinfo(void *, struct accdb_user *, struct accdb_user *);
extern int q_useradd(void *, struct accdb_user *);
extern int q_usermod(void *, struct accdb_user *, struct accdb_user *);
extern int q_userdel(void *, struct accdb_user *);
extern int q_grouptrav(void *, struct accdb_group *);
extern int q_groupinfo(void *, struct accdb_group *, struct accdb_group *);
extern int q_groupadd(void *, struct accdb_group *);
extern int q_groupmod(void *, struct accdb_group *, struct accdb_group *);
extern int q_groupdel(void *, struct accdb_group *);
extern int q_close(void *);
extern int p_sysctl(unsigned int, ...);

static void db_flush__unlock(struct state *);
static int db_flush(struct state *, int);
static long find_next_id(struct state *, long, unsigned long);
static void free_groups(struct state *);
static void free_users(struct state *);
static struct accdb_group *get_group(struct state *, char *);
static struct accdb_user *get_user(struct state *, char *);
static int lock_file(FILE *, enum lock_request);
static void read_config(struct state *);
static int read_groups(struct state *);
static int read_users(struct state *);

MODULE_NAME("Shadow userdb module");
MODULE_DESC("for /etc/passwd and shadow");

const char *DB_passwd = "/etc/passwd", *DB_shadow = "/etc/shadow",
 *DB_group = "/etc/group", *DB_gshadow = "/etc/gshadow";

//-----------------------------------------------------------------------------
int q_open(void **state) {
  struct state *sp;
  int eax;
  if((sp = *state = calloc(1, sizeof(struct state))) == NULL) {
    return -errno;
  }

  read_config(sp);

  if((sp->fp_pw = fopen(DB_passwd, "r+")) == NULL ||
   (sp->fp_sh = fopen(DB_shadow, "r+")) == NULL ||
   (sp->fp_gr = fopen(DB_group, "r+")) == NULL ||
   (sp->fp_gs = fopen(DB_gshadow, "r+")) == NULL) {
    eax = errno;
    fprintf(stderr, "Could not open at least one of the DB core files: %s\n",
     strerror(errno));
    q_close(*state);
    return -eax;
  }

  if((eax = read_users(sp)) < 0 || (eax = read_groups(sp)) < 0) {
    q_close(*state);
    fprintf(stderr, "Error while reading users/groups: %s\n", strerror(-eax));
    return eax;
  }

  sp->utrav = sp->user->first;
  sp->gtrav = sp->group->first;
  return 1;
}

//-----------------------------------------------------------------------------
int q_usertrav(void *state, struct accdb_user *result) {
  struct state *sp = state;
  struct accdb_user *src;
  struct accdb_group qgrp = {gname: NULL};

  if(result == NULL) {
    sp->utrav = sp->user->first;
    return 1;
  }

  if(sp->utrav == NULL) { return 0; }
  src = sp->utrav->ptr;

  while(*src->lname == '+' || *src->lname == '-') {
    // skip NIS users
    if((sp->utrav = sp->utrav->Next) == NULL) { return 0; }
    src = sp->utrav->ptr;
  }

  memcpy(result, src, sizeof(struct accdb_user));
  qgrp.gid = src->gid;
  q_groupinfo(state, &qgrp, &qgrp);
  result->group = qgrp.gname;

  sp->utrav = sp->utrav->Next;
  return 1;
}

int q_userinfo(void *state, struct accdb_user *req, struct accdb_user *dest) {
  struct state *sp = state;
  struct HXlld_node *travp = sp->group->first;

  while(travp != NULL) {
    struct accdb_user *src = travp->ptr;
    if(*src->lname == '+' || *src->lname == '-' || // happily skip NIS users
     (req->lname != NULL && strcmp(req->lname, src->lname) != 0) ||
     (req->uid != -1 && req->uid != src->uid) ||
     (req->gid != -1 && req->gid != src->gid)) {
      travp = travp->Next;
      continue;
    }

    if(dest != NULL) {
      struct accdb_group qgrp = {gid: src->gid, gname: NULL};
      memcpy(dest, src, sizeof(struct accdb_user));
      q_groupinfo(state, &qgrp, &qgrp);
      dest->group = qgrp.gname;
    }
    return 1;
  }

  return 0;
}

int q_useradd(void *state, struct accdb_user *u) {
  struct state *sp = state;
  struct accdb_user *nu;

  if((nu = calloc(1, sizeof(struct accdb_user))) == NULL) { return -errno; }

  /* Four possible group cases:
  1. no group info in U at all
  2. only u->gid available
  3. only u->group available
  4. both available
  In case one, we bomb out, it is useradd's job to get a default group. If
  both are available, the GID is taken. */
  if(u->lname || (u->gid == -1 && u->group == NULL)) {
    errno = EINVAL;
    return 0;
  }

  /* We have to be careful with strings here, as we need to duplicate them for
  us, leave the original ones intact and not have a single foreign pointer in
  our structs. */
  if(u->gid != -1 && u->group != NULL) { // case 4
    struct accdb_group qgrp = {gid: u->gid, gname: u->group};
    if(q_groupinfo(state, &qgrp, &qgrp) <= 0) {
      /* The group which fulfills both requirements (GID _and_ name)
      was not found. */
      errno = ENOENT;
      return 0;
    }
    nu->gid = qgrp.gid;
    nu->group = qgrp.gname;
  } else if(u->gid != -1) { // case 2
    struct accdb_group qgrp = {gid: u->gid, gname: NULL};
    if(q_groupinfo(state, &qgrp, &qgrp) > 0) { nu->group = qgrp.gname; }
    /* qgrp.group might be NULL if qgrp.gid was not found.
    We take the GID anyway. */
    nu->gid = u->gid;
  } else if(u->group != NULL) { // case 3
    struct accdb_group qgrp = {gid: -1, gname: u->group};
    if(q_groupinfo(state, &qgrp, &qgrp) <= 0) {
      errno = ENOENT;
      return 0;
    }
    nu->gid = qgrp.gid;
    nu->group = qgrp.gname;
  }

  nu->lname = HX_strdup(u->lname);
  nu->uid = u->uid;
  if(nu->uid == -1) { nu->uid = find_next_id(sp, -1, ID_USER); }
  nu->gecos = HX_strdup(u->gecos);
  nu->home = HX_strdup(u->home);
  nu->shell = HX_strdup(u->shell);

  nu->pass = HX_strdup(u->pass);
  nu->last_change = u->last_change;
  nu->keep_min = u->keep_min;
  nu->keep_max = u->keep_max;
  nu->warn_age = u->warn_age;
  nu->expiry = u->expiry;
  nu->disabled_since = u->disabled_since;

  // We leave it up to the application to check for duplicates
  HXlld_push(sp->user, nu);
  sp->chg_pw = sp->chg_sh = 1;
  if(db_flush(sp, 0) <= 0) { return -errno; }
  return 1;
}

int q_usermod(void *state, struct accdb_user *u, struct accdb_user *n) {
  struct state *sp = state;
  if(db_flush(sp, 0) <= 0) { return -errno; }
  return 0;
}

int q_userdel(void *state, struct accdb_user *u) {
  struct state *sp = state;
  if(db_flush(sp, 0) <= 0) { return -errno; }
  return 0;
}

//-----------------------------------------------------------------------------
int q_grouptrav(void *state, struct accdb_group *result) {
  struct state *sp = state;
  struct accdb_group *x;

  if(result == NULL) {
    sp->gtrav = sp->group->first;
    return 1;
  }

  if(sp->gtrav == NULL) { return 0; }
  x = sp->gtrav->ptr;

  while(*x->gname == '+' || *x->gname == '-') {
    // skip NIS groups
    if((sp->gtrav = sp->gtrav->Next) == NULL) { return 0; }
    x = sp->gtrav->ptr;
  }

  memcpy(result, x, sizeof(struct accdb_group));
  sp->gtrav = sp->gtrav->Next;
  return 1;
}

int q_groupinfo(void *state, struct accdb_group *r, struct accdb_group *w) {
  struct state *sp = state;
  struct HXlld_node *travp = sp->group->first;

  while(travp != NULL) {
    struct accdb_group *i = travp->ptr;
    if(*i->gname == '+' || *i->gname == '-' || // happily skip NIS groups
     (r->gid != -1 && r->gid != i->gid) ||
     (r->gname != NULL && strcmp(r->gname, i->gname) != 0)) {
      travp = travp->Next;
      continue;
    }

    if(w != NULL) { memcpy(w, i, sizeof(struct accdb_group)); }
    return 1;
  }

  return 0;
}

int q_groupadd(void *state, struct accdb_group *g) {
  struct state *sp = state;
  struct accdb_group *ng;

  if((ng = calloc(1, sizeof(struct accdb_group))) == NULL) { return -errno; }
  if(g->gname == NULL) {
    errno = EINVAL;
    return 0;
  }

  ng->gname = HX_strdup(g->gname);
  ng->gid = g->gid;
  if(ng->gid == -1) { ng->gid = find_next_id(sp, -1, ID_GROUP); }

  HXlld_push(sp->group, ng);
  sp->chg_gr = sp->chg_gs = 1;
  if(db_flush(sp, 0) <= 0) { return -errno; }
  return 1;
}

int q_groupmod(void *state, struct accdb_group *g, struct accdb_group *mask) {
  struct state *sp = state;

  /* We first check if the group in G is available, then make G point to our
  internal data structures. */
  if(q_groupinfo(state, g, NULL) <= 0 ||
   (g = get_group(sp, g->gname)) == NULL) {
    errno = ENOENT;
    return 0;
  }

  if(mask->gname != NULL) { HX_strclone(&g->gname, mask->gname); }
  if(mask->gid != -1) { g->gid = mask->gid; }

  sp->chg_gr = 1;
  if(db_flush(sp, 0) <= 0) { return -errno; }
  return 1;
}

int q_groupdel(void *state, struct accdb_group *g) {
  struct state *sp = state;
  if((g = get_group(sp, g->gname)) == NULL) {
    errno = ENOENT;
    return 0;
  }

  HXlld_del(HXlld_find(sp->group, g));
  free(g->gname);
  free(g);

  sp->chg_gr = sp->chg_gs = 1;
  if(db_flush(sp, 0) <= 0) { return -errno; }
  return 1;
}

int q_close(void *state) {
#define fclose_null(fp) /* guard against NULL */ \
  if((fp) != NULL) { fclose(fp); }
  struct state *sp = state;
  db_flush(sp, 1); // XXX what, if it fails? // YYY: then we're screwed
  fclose_null(sp->fp_pw);
  fclose_null(sp->fp_sh);
  fclose_null(sp->fp_gr);
  fclose_null(sp->fp_gs);
  free_users(sp);
  free_groups(sp);
  free(sp);
  return 1;
#undef fclose_null
}

int p_sysctl(unsigned int req, ...) {
  enum { // accdb_shadow specific requests
    ACCDB_CHDB_PASSWD = 0x4000,
    ACCDB_CHDB_GROUP,
    ACCDB_CHDB_SHADOW,
    ACCDB_CHDB_GSHADOW,
  };

  struct state *sp = NULL;
  int rv = 1;
  va_list argp;

  va_start(argp, req);

  switch(req) {
    case ACCDB_ADDFLAGS:
      // p_sysctl(unsigned int req, struct state *sp, unsigned long fl);
      sp = va_arg(argp, void *);
      sp->flags |= va_arg(argp, unsigned long);
      break;
    case ACCDB_DELFLAGS:
      // p_sysctl(unsigned int req, struct state *sp, unsigned long fl);
      sp = va_arg(argp, void *);
      sp->flags &= ~va_arg(argp, unsigned long);
      break;
    case ACCDB_NEXTUID_SYS:
      // p_sysctl(unsigned int req, struct state *sp);
      rv = find_next_id(va_arg(argp, void *), -1, ID_USER | IDF_SYSTEM);
      break;
    case ACCDB_NEXTUID:
      // p_sysctl(unsigned int req, struct state *sp);
      rv = find_next_id(va_arg(argp, void *), -1, ID_USER);
      break;
    case ACCDB_NEXTGID_SYS:
      // p_sysctl(unsigned int req, struct state *sp);
      rv = find_next_id(va_arg(argp, void *), -1, ID_GROUP | IDF_SYSTEM);
      break;
    case ACCDB_NEXTGID:
      // p_sysctl(unsigned int req, struct state *sp);
      rv = find_next_id(va_arg(argp, void *), -1, ID_GROUP);
      break;
    case ACCDB_FLUSHDB:
      // p_sysctl(unsigned int req, struct state *sp);
      db_flush(va_arg(argp, void *), 1);
      break;
    case ACCDB_LOCK:
      break;
    case ACCDB_UNLOCK:
      break;

    // These internal sysctl requests are only used for debugging
    case ACCDB_CHDB_PASSWD:
      DB_passwd = HX_strdup(va_arg(argp, char *));
      break;
    case ACCDB_CHDB_GROUP:
      DB_group = HX_strdup(va_arg(argp, char *));
      break;
    case ACCDB_CHDB_SHADOW:
      DB_shadow = HX_strdup(va_arg(argp, char *));
      break;
    case ACCDB_CHDB_GSHADOW:
      DB_gshadow = HX_strdup(va_arg(argp, char *));
      break;
    default:
      rv = 0;
      break;
  }

  va_end(argp);
  return rv;
}

//-----------------------------------------------------------------------------
static void db_flush__unlock(struct state *sp) {
  lock_file(sp->fp_pw, LOCK_OFF);
  lock_file(sp->fp_sh, LOCK_OFF);
  lock_file(sp->fp_gr, LOCK_OFF);
  lock_file(sp->fp_gs, LOCK_OFF);
  return;
}

static int db_flush(struct state *sp, int force) {
  if(!force && sp->flags & ACCDB_NOAUTOFLUSH) { return 1; }

  /* Locks the corresponding database files and writes all users stored in the
  state back to disk. */
  if(!lock_file(sp->fp_pw, LOCK_ON) || !lock_file(sp->fp_sh, LOCK_ON) ||
   !lock_file(sp->fp_gr, LOCK_ON) || !lock_file(sp->fp_gs, LOCK_ON)) {
    int se = errno;
    db_flush__unlock(sp);
    return -(errno = se);
  }

  if(sp->chg_pw || sp->chg_sh) { // /etc/passwd and /etc/shadow
    struct HXlld_node *travp = sp->user->first;
    FILE *fpw, *fsh;

    // XXX what if fopen() fails?
    fpw = fopen(DB_passwd, "w");
    fsh = fopen(DB_shadow, "w");
    while(travp != NULL) {
      struct accdb_user *u = travp->ptr;
      char **pr = u->_private;
      fprintf(fpw, "%s:x:%ld:%ld:%s:%s:%s",
       u->lname, u->uid, u->gid, u->gecos, u->home, u->shell);
      if(pr != NULL && pr[0] != NULL) { fprintf(fpw, "%s", pr[0]); }
      fprintf(fpw, "\n");

      fprintf(fsh, "%s:%s:%ld:%ld:%ld:%ld:%ld:%ld",
       u->lname, u->pass, u->last_change, u->keep_min, u->keep_max,
       u->warn_age, u->expiry, u->disabled_since);
      if(pr != NULL && pr[1] != NULL) { fprintf(fsh, "%s", pr[1]); }
      fprintf(fsh, "\n");

      travp = travp->Next;
    }

    fclose(fpw);
    fclose(fsh);
    sp->chg_pw = sp->chg_sh = 0;
  }

  if(sp->chg_gr || sp->chg_gs) { // /etc/group and /etc/gshadow
    struct HXlld_node *travp = sp->group->first;
    FILE *fgr, *fgs;

    fgr = fopen(DB_group, "w");
    fgs = fopen(DB_gshadow, "w");

    while(travp != NULL) {
      struct accdb_group *g = travp->ptr;
      char **pr = g->_private;
      fprintf(fgr, "%s:x:%ld:", g->gname, g->gid);
      if(pr != NULL && pr[0] != NULL) { fprintf(fgr, "%s", pr[0]); }
      fprintf(fgr, "\n");

      fprintf(fgs, "%s:", g->gname);
      if(pr != NULL && pr[1] != NULL) { fprintf(fgs, "%s", pr[1]); }
      else { fprintf(fgs, "*:root:"); }
      fprintf(fgs, "\n");
      travp = travp->Next;
    }

    fclose(fgr);
    fclose(fgs);
    sp->chg_gr = sp->chg_gs = 0;
  }

  db_flush__unlock(sp);
  return 1;
}

static long find_next_id(struct state *sp, long want, unsigned long flags) {
  long min, max;

  if(flags & ID_USER) {
    /* Tries to find the next free UID. If u is >= 0, that specific UID is
    wanted. In that case, u is returned if it is free, or -EEXIST otherwise.
    For u < 0 (just get the next free UID), the next free one is returned or
    -ENOENT if none could be found between sp->uid_min and sp->uid_max. */
    struct HXlld_node *travp = sp->user->first;
    byte used = 0;

    if(flags & IDF_SYSTEM) {
      min = 1;
      max = sp->uid_min - 1;
    } else {
      min = sp->uid_min;
      max = sp->uid_max;
    }

    if(want >= 0) {
      while(travp != NULL) {
        if(((struct accdb_user *)travp->ptr)->uid == want) {
          ++used;
          break;
        }
        travp = travp->Next;
      }
      if(used) { return -(errno = EEXIST); }
      return want;
    }

    /* No specific UID: As the original "useradd" from the Shadow package
    did, pick the UID after the newest user below or equal to UID_MAX. */
    want = min;
    while(travp != NULL) {
      long this = ((struct accdb_user *)travp->ptr)->uid;
      if(this >= want && this <= max) {
        if(want == max) { want = -1; break; }
        want = this + 1;
      }
      travp = travp->Next;
    }

    if(want >= 0) { return want; }

    /* There is no more free UID after the newest user <= UID_MAX, fall back to
    method #2: get the next free UID, starting with UID_MIN. */
    want = min;
    while(want <= max) {
      travp = sp->user->first;
      while(travp != NULL) {
        if(((struct accdb_user *)travp->ptr)->uid == want) {
          ++used;
          break;
        }
        travp = travp->Next;
      }
      if(!used) { return want; }
      --used;
      ++want;
    }

    return -(errno = ENOENT);
  } else if(flags & ID_GROUP) {
    // Same for groups
    struct HXlld_node *travp = sp->group->first;
    byte used = 0;

    if(flags & IDF_SYSTEM) {
      min = 1;
      max = sp->gid_min - 1;
    } else {
      min = sp->gid_min;
      max = sp->gid_max;
    }

    if(want >= 0) {
      while(travp != NULL) {
        if(((struct accdb_user *)travp->ptr)->gid == want) {
          ++used;
          break;
        }
        travp = travp->Next;
      }
      if(used) { return -(errno = EEXIST); }
      return want;
    }

    want = min;
    while(travp != NULL) {
      long this = ((struct accdb_group *)travp->ptr)->gid;
      if(this >= want && this <= max) {
        if(want == max) { want = -1; break; }
        want = this + 1;
      }
      travp = travp->Next;
    }

    if(want >= 0) { return want; }

    want = min;
    while(want <= max) {
      travp = sp->group->first;
      while(travp != NULL) {
        if(((struct accdb_user *)travp->ptr)->uid == want) {
          ++used;
          break;
        }
        travp = travp->Next;
      }
      if(!used) { return want; }
      --used;
      ++want;
    }
    return -(errno = ENOENT);
  }

  return -1;
}

static void free_users(struct state *sp) {
#define free_null(ptr) /* guard against free(NULL) */ \
  if((ptr) != NULL) { free(ptr); }
  struct HXlld_node *travp;
  if(sp->user == NULL) { return; }
  travp = sp->user->first;

  while(travp != NULL) {
    struct accdb_user *u = travp->ptr;
    char **p = u->_private;
    free(u->lname);
    free(u->gecos);
    free(u->home);
    free(u->shell);
    free(u->pass);

    if(p != NULL) {
      free_null(p[0]);
      free_null(p[1]);
      free(p);
    }

    travp = travp->Next;
  }

  HXlld_free(sp->user);
  return;
#undef free_null
}

static void free_groups(struct state *sp) {
#define free_null(ptr) /* guard against free(NULL) */ \
  if((ptr) != NULL) { free(ptr); }
  struct HXlld_node *travp;
  if(sp->group == NULL) { return; }
  travp = sp->group->first;

  while(travp != NULL) {
    struct accdb_group *g = travp->ptr;
    char **p = g->_private;
    free(g->gname);
    if(p != NULL) {
      free_null(p[0]);
      free_null(p[1]);
      free(p);
    }
    travp = travp->Next;
  }

  HXlld_free(sp->group);
  return;
#undef free_null
}

static struct accdb_user *get_user(struct state *sp, char *lname) {
  struct HXlld_node *travp = sp->user->first;

  while(travp != NULL) {
    struct accdb_user *u = travp->ptr;
    if(strcmp(u->lname, lname) == 0) { return u; }
    travp = travp->Next;
  }

  return NULL;
}

static struct accdb_group *get_group(struct state *sp, char *name) {
  // XXX....this looks like a faster function as q_userinfo()...
  // but then again, only compares against names. Keep it?
  struct HXlld_node *travp = sp->group->first;

  while(travp != NULL) {
    struct accdb_group *g = travp->ptr;
    if(strcmp(g->gname, name) == 0) { return g; }
    travp = travp->Next;
  }

  return NULL;
}

static int lock_file(FILE *fp, enum lock_request req) {
  long pos = ftell(fp);
  int rv = -1, se = 0;

  fseek(fp, 0, SEEK_SET);
  fflush(fp);

  switch(req) {
    case LOCK_OFF:
      rv = lockf(fileno(fp), F_ULOCK, 0);
      se = errno;
      break;
    case LOCK_ON:
      rv = lockf(fileno(fp), F_LOCK, 0);
      se = errno;
      break;
  }

  fseek(fp, pos, SEEK_SET);
  fflush(fp);
  errno = se;
  if(rv == -1) { return 0; }
  return 1;
}

static void read_config(struct state *sp) {
  const char *paths[] = {"etc", "/etc", NULL};
  struct rconfig_opt table[] = {
    {"UID_MIN", 'l', &sp->uid_min},
    {"UID_MAX", 'l', &sp->uid_max},
    {"GID_MIN", 'l', &sp->gid_min},
    {"GID_MAX", 'l', &sp->gid_max},
    {NULL, '\0', NULL},
  };

  sp->uid_min = sp->gid_min = 100;
  sp->uid_max = sp->gid_max = 60000;

  accdb_rconfig_pv(paths, "autouid", table, RCONFIG_ONE);
  return;
}

static int read_users(struct state *sp) {
  /* In this function, we read all users into memory. This speeds processing
  up, as seeking and scanning in userinfo() / groupinfo() would take long
  otherwise. We also have it handy in a struct. Doing so also allows us to
  commit all changes to the user database at once (actually only the case with
  the UDB_NOAUTOFLUSH flag).
  
  Strings are allocated by us -- and freed by us. The char * pointers the
  application gets are only for reading (and dup'ing). */
  struct accdb_user *u;
  byte buf[MAXLNLEN];
  if((sp->user = HXlld_init()) == NULL) { return -errno; }

  while(fgets(buf, MAXLNLEN, sp->fp_pw) != NULL) {
    char *bufp = buf, *pa;
    if(*bufp == '#') { continue; }
    if((u = calloc(1, sizeof(struct accdb_user))) == NULL) { break; }
    HX_chomp(bufp);

    u->lname = HX_strdup(HX_strsep(&bufp, ":"));
    HX_strsep(&bufp, ":");
    u->uid = strtol(HX_strsep(&bufp, ":"), NULL, 0);
    u->gid = strtol(HX_strsep(&bufp, ":"), NULL, 0);
    u->gecos = HX_strdup(HX_strsep(&bufp, ":"));
    u->home = HX_strdup(HX_strsep(&bufp, ":"));
    u->shell = HX_strdup(HX_strsep(&bufp, ":"));

    /* Anything after the seven main fields is "reserved for future use",
    however, we need to keep that intact. */
    u->_private = calloc(1, sizeof(void *) * 2);
    if((pa = HX_strsep(&bufp, ":")) != NULL && strlen(pa) > 0) {
      ((char **)u->_private)[0] = HX_strdup(pa);
    }

    /* Stuff is pushed in a linked list. The list is unordered, so it has the
    same order as /etc/passwd. This is intentional (and suffices). */
    HXlld_push(sp->user, u);
  }

  // Update the sp->user table with the data from /etc/shadow
  while(fgets(buf, MAXLNLEN, sp->fp_sh) != NULL) {
    char *bufp = buf, *lname, *pa;
    if(*buf == '#') { continue; }
    HX_chomp(buf);

    lname = HX_strsep(&bufp, ":");
    if((u = get_user(sp, lname)) == NULL) { continue; } // orphaned entry
    u->pass = HX_strdup(HX_strsep(&bufp, ":"));
    u->last_change = strtol(HX_strsep(&bufp, ":"), NULL, 0);
    u->keep_min = strtol(HX_strsep(&bufp, ":"), NULL, 0);
    u->keep_max = strtol(HX_strsep(&bufp, ":"), NULL, 0);
    u->warn_age = strtol(HX_strsep(&bufp, ":"), NULL, 0);
    u->expiry = strtol(HX_strsep(&bufp, ":"), NULL, 0);
    u->disabled_since = strtol(HX_strsep(&bufp, ":"), NULL, 0);

    if((pa = HX_strsep(&bufp, ":")) != NULL && strlen(pa) > 0) {
      ((char **)u->_private)[1] = HX_strdup(pa);
    }
  }

  return 1;
}

static int read_groups(struct state *sp) {
  // Comments to be found in read_users()
  struct accdb_group *g;
  byte buf[MAXLNLEN];
  if((sp->group = HXlld_init()) == NULL) { return -errno; }

  while(fgets(buf, MAXLNLEN, sp->fp_gr) != NULL) {
    char *bufp = buf, *pa;
    if(*bufp == '#') { continue; }
    if((g = malloc(sizeof(struct accdb_group))) == NULL) { return -errno; }
    HX_chomp(bufp);

    g->gname = HX_strdup(HX_strsep(&bufp, ":"));
    // second field is to be ignored, it is located in gshadow
    HX_strsep(&bufp, ":");
    g->gid = strtol(HX_strsep(&bufp, ":"), NULL, 0);
    g->_private = calloc(1, sizeof(void *) * 2);

    if((pa = HX_strsep(&bufp, ":")) != NULL && strlen(pa) > 0) {
      ((char **)g->_private)[0] = HX_strdup(pa);
    }

    HXlld_push(sp->group, g);
  }

  // Update the sp->group table with the data from /etc/gshadow
  while(fgets(buf, MAXLNLEN, sp->fp_gs) != NULL) {
    char *bufp = buf, *name;
    if(*bufp == '#') { continue; }
    HX_chomp(bufp);

    name = HX_strsep(&bufp, ":");
    if((g = get_group(sp, name)) == NULL) { continue; } // orphaned entry
    if(strlen(bufp) > 0) { ((char **)g->_private)[1] = HX_strdup(bufp); }
  }

  return 1;
}

//==[ End of file ]============================================================
