/*=============================================================================
vitalnix/accdb - User Management Suite - Unified Account Database
groupdel.c - Delete a group
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <string.h>
#include <accdb.h>

typedef unsigned char byte;

enum {
  E_SUCCESS = 0,
  E_OTHER,       // other error, see errno
  E_OPEN,        // unable to open UDB module or DB
  E_NOEXIST,     // group does not exist
  E_UPDATE,      // db->groupdel() did not return ok
  E_CLOSE,       // db->close() did not return ok
};

struct {
  char *gname;
  int force;
} opt;

static int get_options(int *, const char ***);
static int count_args(const char **);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
  struct accdb_module *db;
  struct accdb_group mmask;
  int eax, rv = E_SUCCESS;

  if(!(eax = get_options(&argc, (const char ***)&argv))) { return E_OTHER; }

  if((db = accdb_load(module_path)) == NULL) {
    fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
     module_path, strerror(errno));
    return E_OPEN;
  }

  if((eax = db->open(&db->state)) <= 0) {
    fprintf(stderr, "Could not open ACCDB back-end: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_OPEN;
    goto __main__close_accdb;
  }

  memset(&mmask, 0, sizeof(struct accdb_group));
  mmask.gid = -1;
  mmask.gname = opt.gname = *argv;

  if((eax = db->groupinfo(db->state, &mmask, NULL)) < 0) {
    fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_OTHER;
    goto __main__close_backend;
  }

  if(eax == 0) {
    fprintf(stderr, "Group \"%s\" does not exist\n", mmask.gname);
    rv = E_NOEXIST;
    goto __main__close_backend;
  }

  //check_pri_group(db, &mmask);

  if((eax = db->groupdel(db->state, &mmask)) <= 0) {
    fprintf(stderr, "Error: Deleting group failed: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_UPDATE;
  }

__main__close_backend:
  if((eax = db->close(db->state)) <= 0) {
    fprintf(stderr, "Warning: DB was not cleanly closed:"
     " %s (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
    if(rv == E_SUCCESS) { rv = E_CLOSE; }
  }

__main__close_accdb:
  accdb_unload(db);
  return rv;
}

//-----------------------------------------------------------------------------
static int count_args(const char **args) {
  int argk = 0;
  while(*args++ != NULL) { ++argk; }
  return argk;
}

static int get_options(int *argc, const char ***argv) {
  static const char *_empty_argv[] = {NULL};
  const struct poptOption options_table[] = {
    // New, uudb-userdel options
    {NULL, 'F', POPT_ARG_NONE, &opt.force, '\0',
     "Force deletion of a group even if users have it as primary group", NULL},
    {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
     "Use a different module than \"*\" (the default)", "name"},
    POPT_AUTOHELP
    POPT_TABLEEND
  };

  poptContext ctx;
  const char **args;
  int c, argk;

  ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
  poptSetOtherOptionHelp(ctx, "[OPTIONS...] group");
  while((c = poptGetNextOpt(ctx)) >= 0);
  if(c < -1) {
    fprintf(stderr, "%s: %s: %s\n", **argv, poptBadOption(ctx, 0),
     poptStrerror(c));
    return 0;
  }

  if((args = poptGetArgs(ctx)) != NULL) {
    argk = count_args(args);
    poptDupArgv(argk, args, argc, argv);
  } else {
    *argc = 0;
    *argv = _empty_argv;
  }

  if(args == NULL || *args == NULL) {
    // gname is mandatory
    poptPrintUsage(ctx, stderr, 0);
    return 0;
  }

  poptFreeContext(ctx);
  return 1;
}

//==[ End of file ]============================================================
